{% extends "layout.html" %}
{% block body %}
<div class="deck">

    {% if version == "3.2" or version == "4.0" %}
        <p class="developmentversion">
        This document is for Celery's development version, which can be
        significantly different from previous releases. Get old docs here:

        <a href="http://docs.celeryproject.org/en/latest/{{ pagename }}{{ file_suffix }}">3.0</a>.
        </p>
        {% else %}
        <p>
        This document describes the current stable version of Celery ({{ version }}). For development docs,
        <a href="http://docs.celeryproject.org/en/master/{{ pagename }}{{ file_suffix }}">go here</a>.
        </p>
    {% endif %}

</div>
    {{ body }}
{% endblock %}
